/* ********************************************************
    clique.cpp

    Author = Hideyuki Harada
    Nihon Univ. Tani Research Group


    Eguchi's Algorithm for resolving MaxClique problem
    designed for "g++ 3.3.3" with "LEDA 4.5"
   ******************************************************** */


#include<LEDA/graph.h>
#include<LEDA/list.h>
#include<time.h>
#include<iostream.h>
#include<fstream.h>
#include<string.h>
#include<LEDA/d_array.h>

#define P_VERSION "clique 0.02"

using namespace leda;

graph G;           // undirected graph
list<node> V;
list<edge> E;


list<node> Qmax;   // node set for MaxClique
list<node> Q;

int Branch_count;  // number of counted SST branch

int coloring_count;

void read_graph(int argc, char *argv[]);   // read filename from command line

void initialize(list<node> *R);  // clear Q and put V to R

void show_result(float T, char *argv[]);  // write result

void newexpand(list<node> R, d_array<node, int> N);  // basic algorithm

void CLIQUE(void);

int coloring(list<node> *Rp, d_array<node, int> *N);

void sort_by_degree(list<node> *R);

int cmp_deg(const node &v, const node &u);

void check_degree(node v);

//*************************************************


int main(int argc, char *argv[]){

  float T = used_time();

  if(argc != 2){
    cout << "usage: clique [file]" << endl;
    exit(1);
  }

  read_graph(argc, argv);

  cout << "resolving " << argv[1] << " ..." << endl;

  // start procedure
  CLIQUE();

  show_result(T, argv);

  cout << "done." << endl;

  return 0;

}

void CLIQUE(void){

  list<node> R;
  d_array<node, int> N;
  int c,i,a;
  node v;


  initialize(&R);
  sort_by_degree(&R);

  c = degree(R.head());
  for(i = 0 ; i < c; i++){
    v = R[R[i]];
    N[v] = i+1;
  }

  a = R.size();
  for(; i < a; i++){
    v = R[R[i]];
    N[v] = c+1;
  }

  newexpand(R, N);

}

void newexpand(list<node> R, d_array<node, int> N){

  node p, v;
  list<node> Rp;
  d_array<node, int> M;
  int c;

  while(!R.empty()){

    Branch_count++;
    p = R.tail();

    if(Q.size() + N[p] > Qmax.size()){
      Q.append(p);

      Rp.clear();
      // Rp = intersect(R, G.adj_nodes(p));
      forall_adj_nodes(v, p){
	if(R.search(v) != NULL && degree(v) > Qmax.size()){
	  Rp.append(v);
	}
      }
 
      if(!Rp.empty()){
	M.clear();
	c = coloring(&Rp, &M);
        newexpand(Rp, M);
      }
      else{
	if(Q.size() > Qmax.size()){  // apply new MaxClique
	  Qmax.clear();
	  forall(v, Q){
	    Qmax.append(v);
	  }
	  forall_nodes(v, G){
	    if(degree(v) <= Qmax.size()){
	      check_degree(v);
	    }
	  }
	}
      }
      Q.remove(p);
    }
    else{
      return;
    }
    R.remove(p);
  }
  return;

}

void check_degree(node v){

  node u;
  list<node> N;
  
  N = G.adj_nodes(v);
  printf("b%d\n", degree(N.head()));
  G.hide_node(v);
  printf("a%d\n", degree(N.head()));

  forall(u, N){
    if(degree(u) <= Qmax.size()){
      check_degree(u);
    }
  }
  
  return;

}

int coloring(list<node> *Rp, d_array<node, int> *N){

  int i, c, k, flag;
  list<node> C[500];
  node p, v;

  c=0;

  while(!((*Rp).empty())){

    p = (*Rp).head();

    k=0;
    do{
      flag = 0;
      k++;
      forall(v, C[k]){
	coloring_count++;
	if((G.adj_nodes(p)).search(v) != NULL){
	  flag = 1;
	  break;
	}
      }
    }while(flag == 1);
    
    if(k > c){
      c = k;
    }
    C[k].append(p);
    (*N)[p] = k;
    
    (*Rp).remove(p);
  }
  // color(v) = min{}

  for(i = 1; i <= c; i++){
    forall(p, C[i]){
      (*Rp).append(p);
    }
  }

  return c;
}


void sort_by_degree(list<node> *R){
  

  (*R).sort(*cmp_deg, 0);


  return;

}

int cmp_deg(const node &v, const node &u){

  if(degree(v) < degree(u)){
    return 1;
  }

  else
    if(degree(v) == degree(u)){
      return 0;
    }

    else
      return -1;
}

void read_graph(int argc, char *argv[]){
  node v;

  G.read(argv[1]);
  G.make_undirected();
  forall_nodes(v, G){
    if(degree(v) == 0)
      G.del_node(v);
  }
  E = G.all_edges();
  V = G.all_nodes();
  

  return;
}

void initialize(list<node> *R){

  Q.clear();
  (*R) = G.all_nodes();
  Qmax.clear();
  Branch_count = 0;
  coloring_count =0;

  return;
}

void show_result(float T, char *argv[]){

  node v;
  ofstream fout;
  time_t t=time(NULL);
  string ofile;

  G.restore_all_nodes();

  ofile = ".result";
 
  fout.open(argv[1]+ofile, ios::app);

  fout << asctime(localtime(&t));
  fout << "Program version = " << P_VERSION << endl;  
  fout << "Filename = " << argv[1] << endl;
  fout << " Number of vertexes = " << G.number_of_nodes() << endl;
  fout << " Number of edges = " << G.number_of_edges() << endl;
  fout << "Used time = " << used_time(T) << endl;
  fout << "SST Branch count = " << Branch_count << endl;
  fout << "coloring node count = " << coloring_count << endl;
  fout << "MAX CLIQUE SIZE = " << Qmax.size() << endl;
  fout << "CLIQUE ELEMENT(S) :" << endl;

  forall(v, Qmax){
    fout << " " << index(v);
  }
  fout << endl;

  fout << endl;

  fout.close();
 
  return;
}
